#!/usr/bin/env python
# coding=utf-8
'''
Author: Liu Kun && 16031215@qq.com
Date: 2024-09-17 16:55:11
LastEditors: Liu Kun && 16031215@qq.com
LastEditTime: 2024-10-06 19:34:57
FilePath: \\Python\\My_Funcs\\OAFuncs\\OAFuncs\\oa_cmap.py
Description:  
EditPlatform: vscode
ComputerInfo: XPS 15 9510
SystemInfo: Windows 11
Python Version: 3.11
'''



import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

__all__ = ['show', 'extract_colors', 'create_custom',
           'create_diverging', 'create_5rgb_txt']

# ** 将cmap用填色图可视化（官网摘抄函数）
def show(colormaps: list):
    """
    Helper function to plot data with associated colormap.
    example:
    cmap = ListedColormap(["darkorange", "gold", "lawngreen", "lightseagreen"])
    show([cmap])
    """
    np.random.seed(19680801)
    data = np.random.randn(30, 30)
    n = len(colormaps)
    fig, axs = plt.subplots(1, n, figsize=(n * 2 + 2, 3),
                            constrained_layout=True, squeeze=False)
    for [ax, cmap] in zip(axs.flat, colormaps):
        psm = ax.pcolormesh(data, cmap=cmap, rasterized=True, vmin=-4, vmax=4)
        fig.colorbar(psm, ax=ax)
    plt.show()


# ** 将cmap转为list，即多个颜色的列表
def extract_colors(cmap, n=256):
    '''
    cmap        : cmap名称
    n           : 提取颜色数量
    return      : 提取的颜色列表
    example     : out_cmap = extract_colors('viridis', 256)
    '''
    c_map = mpl.colormaps.get_cmap(cmap)
    out_cmap = [c_map(i) for i in np.linspace(0, 1, n)]
    return out_cmap


# ** 自制cmap，多色，可带位置
def create_custom(colors: list, nodes=None):  # 利用颜色快速配色
    '''
    func        : 自制cmap，自动确定颜色位置（等比例）
    description : colors可以是颜色名称，也可以是十六进制颜色代码
    param        {*} colors 颜色
    param        {*} nodes 颜色位置，默认不提供，等间距
    return       {*} c_map
    example     : c_map = mk_cmap(['#C2B7F3','#B3BBF2','#B0CBF1','#ACDCF0','#A8EEED'])
                c_map = mk_cmap(['aliceblue','skyblue','deepskyblue'],[0.0,0.5,1.0])
    '''
    if nodes is None:  # 采取自动分配比例
        cmap_color = mpl.colors.LinearSegmentedColormap.from_list(
            'mycmap', colors)
    else:  # 按照提供比例分配
        cmap_color = mpl.colors.LinearSegmentedColormap.from_list(
            "mycmap", list(zip(nodes, colors)))
    return cmap_color

# ** 自制diverging型cmap，默认中间为白色


def create_diverging(colors: list):
    '''
    func        : 自制cmap，双色，中间默认为白色；如果输入偶数个颜色，则中间为白，如果奇数个颜色，则中间色为中间色
    description : colors可以是颜色名称，也可以是十六进制颜色代码
    param        {*} colors
    return       {*}
    example     : diverging_cmap = mk_cmap_diverging(
    ["#4e00b3", "#0000FF", "#00c0ff", "#a1d3ff", "#DCDCDC", "#FFD39B", "#FF8247", "#FF0000", "#FF5F9E"])
    '''
    # 自定义颜色位置
    n = len(colors)
    nodes = np.linspace(0.0, 1.0, n + 1 if n % 2 == 0 else n)
    newcolors = colors
    if n % 2 == 0:
        newcolors.insert(int(n / 2), '#ffffff')  # 偶数个颜色，中间为白色
    cmap_color = mpl.colors.LinearSegmentedColormap.from_list(
        "mycmap", list(zip(nodes, newcolors)))
    return cmap_color

# ** 根据RGB的txt文档制作色卡（利用Grads调色盘）


def create_5rgb_txt(rgb_txt_filepath: str):  # 根据RGB的txt文档制作色卡/根据rgb值制作
    '''
    func        : 根据RGB的txt文档制作色卡
    description : rgb_txt_filepath='E:/python/colorbar/test.txt'
    param        {*} rgb_txt_filepath txt文件路径
    return       {*} camp
    example     : cmap_color=dcmap(path)
    '''
    with open(rgb_txt_filepath) as fid:
        data = fid.readlines()
    n = len(data)
    rgb = np.zeros((n, 3))
    for i in np.arange(n):
        rgb[i][0] = data[i].split(',')[0]
        rgb[i][1] = data[i].split(',')[1]
        rgb[i][2] = data[i].split(',')[2]
    max_rgb = np.max(rgb)
    if max_rgb > 2:  # 如果rgb值大于2，则认为是0-255的值，需要归一化
        rgb = rgb / 255.0
    icmap = mpl.colors.ListedColormap(rgb, name='my_color')
    return icmap




if __name__ == '__main__':
    # ** 测试自制cmap
    colors = ['#C2B7F3', '#B3BBF2', '#B0CBF1', '#ACDCF0', '#A8EEED']
    nodes = [0.0, 0.2, 0.4, 0.6, 1.0]
    c_map = create_custom(colors, nodes)
    show([c_map])

    # ** 测试自制diverging型cmap
    diverging_cmap = create_diverging(["#4e00b3", "#0000FF", "#00c0ff",
                                       "#a1d3ff", "#DCDCDC", "#FFD39B", "#FF8247", "#FF0000", "#FF5F9E"])
    show([diverging_cmap])

    # ** 测试根据RGB的txt文档制作色卡
    file_path = 'E:/python/colorbar/test.txt'
    cmap_color = create_5rgb_txt(file_path)

    # ** 测试将cmap转为list
    out_cmap = extract_colors('viridis', 256)



#!/usr/bin/env python
# coding=utf-8
'''
Author: Liu Kun && 16031215@qq.com
Date: 2024-09-17 15:07:13
LastEditors: Liu Kun && 16031215@qq.com
LastEditTime: 2024-10-06 19:00:51
FilePath: \\Python\\My_Funcs\\OAFuncs\\OAFuncs\\oa_file.py
Description:  
EditPlatform: vscode
ComputerInfo: XPS 15 9510
SystemInfo: Windows 11
Python Version: 3.11
'''



import glob
import os
import re
import shutil

__all__ = ['link_file', 'copy_file', 'rename_files', 'make_folder',
           'clear_folder', 'remove_empty_folders', 'remove']

def link_file(src_pattern, dst):
    '''
    # 使用示例
    # link_file(r'/data/hejx/liukun/era5/*', r'/data/hejx/liukun/Test/')
    # link_file(r'/data/hejx/liukun/era5/py.o*', r'/data/hejx/liukun/Test/py.o')
    # link_file(r'/data/hejx/liukun/era5/py.o*', r'/data/hejx/liukun/Test')
    '''
    src_pattern = str(src_pattern)
    # 使用glob.glob来处理可能包含通配符的src
    src_files = glob.glob(src_pattern)
    if not src_files:
        raise FileNotFoundError('文件不存在: {}'.format(src_pattern))

    # 判断dst是路径还是包含文件名的路径
    if os.path.isdir(dst):
        # 如果dst是路径，则保持源文件的文件名
        dst_dir = dst
        for src_file in src_files:
            src_file_basename = os.path.basename(src_file)
            dst_file = os.path.join(dst_dir, src_file_basename)
            if os.path.exists(dst_file):
                os.remove(dst_file)
            os.symlink(src_file, dst_file)
            print(f'创建符号链接: {src_file} -> {dst_file}')
    else:
        # 如果dst包含文件名，则创建链接后重命名
        dst_dir = os.path.dirname(dst)
        os.makedirs(dst_dir, exist_ok=True)
        # 只处理第一个匹配的文件
        src_file = src_files[0]
        dst_file = dst
        if os.path.exists(dst_file):
            os.remove(dst_file)
        os.symlink(src_file, dst_file)
        print(f'创建符号链接并重命名: {src_file} -> {dst_file}')


def copy_file(src_pattern, dst):
    '''
    # 使用示例
    # copy_file(r'/data/hejx/liukun/era5/py.o*', r'/data/hejx/liukun/Test/py.o')
    # link_file(r'/data/hejx/liukun/era5/*', r'/data/hejx/liukun/Test/')
    # link_file(r'/data/hejx/liukun/era5/py.o*', r'/data/hejx/liukun/Test/py.o')
    # copy_file(r'/data/hejx/liukun/era5/py.o*', r'/data/hejx/liukun/Test')
    '''
    src_pattern = str(src_pattern)
    # 使用glob.glob来处理可能包含通配符的src
    src_files = glob.glob(src_pattern)
    if not src_files:
        raise FileNotFoundError('文件不存在: {}'.format(src_pattern))

    # 判断dst是路径还是包含文件名的路径
    if os.path.isdir(dst):
        # 如果dst是路径，则保持源文件的文件名
        dst_dir = dst
        for src_file in src_files:
            src_file_basename = os.path.basename(src_file)
            dst_file = os.path.join(dst_dir, src_file_basename)
            if os.path.exists(dst_file):
                if os.path.isdir(dst_file):
                    shutil.rmtree(dst_file)
                else:
                    os.remove(dst_file)
            if os.path.isdir(src_file):
                shutil.copytree(src_file, dst_file, symlinks=True)
            else:
                shutil.copy2(src_file, dst_file)
            print(f'复制文件或目录: {src_file} -> {dst_file}')
    else:
        # 如果dst包含文件名，则复制后重命名
        dst_dir = os.path.dirname(dst)
        os.makedirs(dst_dir, exist_ok=True)
        # 只处理第一个匹配的文件
        src_file = src_files[0]
        dst_file = dst
        if os.path.exists(dst_file):
            if os.path.isdir(dst_file):
                shutil.rmtree(dst_file)
            else:
                os.remove(dst_file)
        if os.path.isdir(src_file):
            shutil.copytree(src_file, dst_file, symlinks=True)
        else:
            shutil.copy2(src_file, dst_file)
        print(f'复制文件或目录并重命名: {src_file} -> {dst_file}')


def rename_files(directory, old_str, new_str):
    '''
    # 使用示例
    directory_path = r"E:\Code\Matlab\Master\Ocean\ROMS\CROCO-1.3.1\My_Models\windfarm\CROCO_FILES"
    old_str = "croco"
    new_str = "roms"
    rename_files(directory_path, old_str, new_str)
    '''
    # 获取目录下的所有文件
    files = os.listdir(directory)

    # 构建正则表达式以匹配要替换的字符串
    pattern = re.compile(re.escape(old_str))

    # 遍历目录下的文件
    for filename in files:
        # 检查文件名中是否包含要替换的字符串
        if pattern.search(filename):
            # 构建新的文件名
            new_filename = pattern.sub(new_str, filename)

            # 构建旧文件的完整路径
            old_path = os.path.join(directory, filename)

            # 构建新文件的完整路径
            new_path = os.path.join(directory, new_filename)

            # 重命名文件
            os.rename(old_path, new_path)
            print(f"重命名文件: {old_path} -> {new_path}")


# ** 创建子文件夹（可选清空）
def make_folder(rootpath: str, folder_name: str, clear=0) -> str:
    folder_path = os.path.join(str(rootpath), str(folder_name))
    if clear:
        shutil.rmtree(folder_path, ignore_errors=True)
    os.makedirs(folder_path, exist_ok=True)
    return folder_path

# ** 清空文件夹
def clear_folder(folder_path):
    folder_path = str(folder_path)
    if os.path.exists(folder_path):
        try:
            # 遍历文件夹中的所有文件和子文件夹
            for filename in os.listdir(folder_path):
                file_path = os.path.join(folder_path, filename)
                # 判断是文件还是文件夹
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)  # 删除文件或链接
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)  # 删除子文件夹
            print(f'成功清空文件夹: {folder_path}')
        except Exception as e:
            print(f'清空文件夹失败: {folder_path}')
            print(e)


# ** 清理空文件夹
def remove_empty_folders(path, print_info=1):
    path = str(path)
    # 遍历当前目录下的所有文件夹和文件
    for root, dirs, files in os.walk(path, topdown=False):
        # 遍历文件夹列表
        for folder in dirs:
            folder_path = os.path.join(root, folder)
            # 判断文件是否有权限访问
            try:
                os.listdir(folder_path)
            except OSError:
                continue
            # 判断文件夹是否为空
            if not os.listdir(folder_path):
                # 删除空文件夹
                try:
                    os.rmdir(folder_path)
                    print(f"Deleted empty folder: {folder_path}")
                except OSError:
                    if print_info:
                        print(f"Skipping protected folder: {folder_path}")
                    pass


# ** 删除相关文件，可使用通配符
def remove(pattern):
    '''
    remove(r'E:\Code\Python\Model\WRF\Radar2\bzip2-radar-0*')
    # or
    os.chdir(r'E:\Code\Python\Model\WRF\Radar2')
    remove('bzip2-radar-0*')
    '''
    # 使用glob.glob来获取所有匹配的文件
    # 可以使用通配符*来匹配所有文件
    pattern = str(pattern)
    file_list = glob.glob(pattern)
    for file_path in file_list:
        if os.path.exists(file_path):
            try:
                shutil.rmtree(file_path)
                print(f'成功删除文件: {file_path}')
            except Exception as e:
                print(f'删除文件失败: {file_path}')
                print(e)
        else:
            print(f'文件不存在: {file_path}')


if __name__ == '__main__':
    # newpath = make_folder('D:/Data/2024/09/17/', 'var1', clear=1)
    # print(newpath)
    pass

    remove(r'I:\Delete\test\*')



#!/usr/bin/env python
# coding=utf-8
'''
Author: Liu Kun && 16031215@qq.com
Date: 2024-09-17 17:26:11
LastEditors: Liu Kun && 16031215@qq.com
LastEditTime: 2024-10-06 19:10:45
FilePath: \\Python\\My_Funcs\\OAFuncs\\OAFuncs\\oa_draw.py
Description:  
EditPlatform: vscode
ComputerInfo: XPS 15 9510
SystemInfo: Windows 11
Python Version: 3.11
'''



import warnings
from mpl_toolkits.axes_grid1 import make_axes_locatable
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter
import cartopy.feature as cfeature
import cartopy.crs as ccrs
import math
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

__all__ = ['create_gif', 'xy2lonlat', 'plot_contourf',
           'plot_contourf_lonlat', 'plot_quiver', 'plot_contourf_cartopy']

warnings.filterwarnings('ignore')



# ** 将生成图片/已有图片制作成动图
def create_gif(image_list: list, gif_name: str, duration=0.2):  # 制作动图，默认间隔0.2
    '''
    func        : 制作动图，将已有图片拼接
    description : Gif格式动图
    param        {*} image_list 图片列表
    param        {*} gif_name 动图名称（含路径）
    param        {*} duration 动图间隔
    return       {*} 自动保存至指定路径（包含于动图名称中）
    example     : 
    '''
    import imageio.v2 as imageio
    frames = []
    for image_name in image_list:
        frames.append(imageio.imread(image_name))
    imageio.mimsave(gif_name, frames, format='GIF', duration=duration)
    print('Gif制作完成！')
    return


# ** 转化经/纬度刻度
def xy2lonlat(xy, lonlat='lon', decimal=2):
    def format_longitude(x_list):
        out_list = []
        for x in x_list:
            if x > 180:
                x -= 360
            # degrees = int(abs(x))
            degrees = round(abs(x), decimal)
            direction = "E" if x >= 0 else "W"
            out_list.append(
                f"{degrees:.{decimal}f}°{direction}" if x != 0 and x != 180 else f"{degrees}°")
        return out_list if len(out_list) > 1 else out_list[0]

    def format_latitude(y_list):
        out_list = []
        for y in y_list:
            if y > 90:
                y -= 180
            # degrees = int(abs(y))
            degrees = round(abs(y), decimal)
            direction = "N" if y >= 0 else "S"
            out_list.append(
                f"{degrees:.{decimal}f}°{direction}" if y != 0 else f"{degrees}°")
        return out_list if len(out_list) > 1 else out_list[0]

    if lonlat == 'lon':
        return format_longitude(xy)
    elif lonlat == 'lat':
        return format_latitude(xy)


# ** 设置colorbar格式
class _MyFormatter(mpl.ticker.ScalarFormatter):
    def __init__(self, cticks, fmt=None, useOffset=True, useMathText=True):
        mpl.ticker.ScalarFormatter.__init__(self, useOffset, useMathText)
        self.cticks = cticks
        self.fmt = fmt
        self.vmin = min(cticks)
        self.vmax = max(cticks)
        self.p_n = self.vmin < 0 and self.vmax > 0
        self.magnitude_min = int(
            math.modf(math.log10(min(abs(cticks[cticks != 0]))))[1])
        self.magnitude_max = int(math.modf(math.log10(max(abs(cticks))))[1])
        # print(self.vmin, self.vmax)

    def __call__(self, x, pos):
        if ((abs(x) < 1e-2) or (abs(x) > 1e4)) and x != 0:
            # if self.magnitude_max - self.magnitude_min == 1 and (int(math.modf(math.log10(abs(x)))[1]) == self.magnitude_min):
            #     a, b = '{:.1e}'.format(x).split('e')
            #     b = int(b)
            #     return '${}{:.2f} \\times 10^{{{}}}$'.format(' ' if self.p_n and x > 0 else '', float(a)/10, b+1)
            # else:
            #     return '${}{} \\times 10^{{{}}}$'.format(' ' if self.p_n and x > 0 else '', *'{:.2e}'.format(x).split('e'))
            if self.magnitude_max - self.magnitude_min == 1 and (int(math.modf(math.log10(abs(x)))[1]) == self.magnitude_min):
                a, b = '{:.1e}'.format(x).split('e')
                a = float(a) / 10
                b = int(b) + 1
            else:
                a, b = '{:.2e}'.format(x).split('e')
                a = float(a)
                b = int(b)
            # return '${}{} \\times 10^{{{}}}$'.format(' ' if (self.p_n and x > 0) else '', a, b)
            return '${}{:.2f}$'.format(' ' if (self.p_n and x > 0) else '', a)
        elif x == 0:
            return '0'
        else:
            return mpl.ticker.ScalarFormatter.__call__(self, x, pos)


# ** 绘制单张填色图
def plot_contourf(pic_data, picname=None, c_map='rainbow', minmax=None, labels=None, ticks_space=None, ticks=None, figsize=(12, 9)):
    '''
    func        : 绘制填色等值线图，单张
    description : 绘制单张填色等值线图，输入参数为横纵坐标、等值数据、图例标注等
    param       {*}pic_data : 填色等值线图的等值数据
    param       {*}picname : 图片保存的文件名(含路径)
    param       {*}c_map    : 颜色映射，默认rainbow
    param       {*}minmax   : 指定绘图的最大、小值，默认不指定
    param       {*}labels   : x、y轴以及图例的标注，默认不标注
    param       {*}ticks_space : x、y轴刻度，以及所显示的标签，默认不显示
    param       {*}figsize  : 图片大小，默认(12,9)
    example     : plot_contourf(pic_data, pictpath, var_name, c_map='bwr', labels=None, ticks_space=None, ticks=None, h=0, figsize=(12, 9))
    '''
    cmap = mpl.colormaps.get_cmap(c_map)
    if minmax is not None:
        value_min, value_max = minmax[0], minmax[1]
    else:
        value_min, value_max = pic_data.nanmin(), pic_data.nanmax()
    v_bry = max(abs(value_min), abs(value_max))
    flag = (value_min < 0) and (value_max > 0)
    norm = mpl.colors.TwoSlopeNorm(
        vmin=-1 * v_bry, vcenter=0, vmax=v_bry) if flag else mpl.colors.Normalize(vmin=value_min, vmax=value_max)
    cticks = [num for num in np.linspace(-1 * v_bry if flag else value_min,
                                         v_bry if flag else value_max, 9)] if value_min != value_max else None
    levels = np.linspace(-1 * v_bry, v_bry,
                         20) if flag else None if value_min == value_max else np.linspace(value_min, value_max, 20)

    shape = np.array(pic_data).shape
    x, y = np.meshgrid(np.arange(shape[1]), np.arange(shape[0]))

    fig, ax = plt.subplots(figsize=figsize)
    flag_lc = levels is not None and cticks is not None
    CS = ax.contourf(x, y, pic_data, cmap=cmap, norm=norm, levels=levels, extend='both') if flag_lc else ax.contourf(
        x, y, pic_data, cmap=cmap, norm=norm, extend='both')
    cb = fig.colorbar(CS, ax=ax, orientation='vertical', shrink=1, format='%.3g', spacing='uniform', ticks=cticks) if cticks is not None else fig.colorbar(
        CS, ax=ax, orientation='vertical', shrink=1, format='%.3g', spacing='uniform')
    """%.3g采用的是自动调整格式，也可设置为%.3f，则改为3位小数"""

    # 将格式化器设置为自定义的函数
    class_cticks = np.array(cb.get_ticks())
    # print(class_cticks)
    cct = abs(class_cticks[class_cticks != 0])
    if (min(cct) < 1e-2) or max(cct) > 1e4:  # 判断是否需要采用科学计数法
        cb.formatter = _MyFormatter(class_cticks)
        cb.update_ticks()

    if labels is not None:
        cb.set_label(labels['c'])
        plt.xlabel(labels['x'])
        plt.ylabel(labels['y'])
    if ticks_space is not None and ticks is not None:
        plt.xticks(np.arange(0, len(x[0, :]) +
                   1e-5, ticks_space['x']), ticks['x'])
        plt.yticks(np.arange(0, len(y[:, 0]) +
                   1e-5, ticks_space['y']), ticks['y'])

    plt.title('Min: {:.3g}, Max: {:.3g}'.format(
        pic_data.min(), pic_data.max()))

    plt.savefig(
        picname, bbox_inches='tight') if picname is not None else plt.show()
    # plt.show()
    plt.clf()
    # 关闭当前figure
    plt.close()


# ** 画等高线图，带经纬度坐标轴
def plot_contourf_lonlat(data, lon, lat, interval=5, picname=None, c_map='rainbow'):
    if len(lon.shape) == 2:
        lon = lon[0, :]
    if len(lat.shape) == 2:
        lat = lat[:, 0]
    # interval是经纬度间隔，单位为°
    # 将lon，lat作为坐标轴刻度显示

    def format_longitude(x):
        if x > 180:
            x -= 360
        degrees = int(abs(x))
        direction = "E" if x >= 0 else "W"
        return f"{degrees}°{direction}" if x != 0 and x != 180 else f"{degrees}°"

    def format_latitude(y):
        if y > 90:
            y -= 180
        degrees = int(abs(y))
        direction = "N" if y >= 0 else "S"
        return f"{degrees}°{direction}" if y != 0 else f"{degrees}°"

    plt.contourf(data, cmap=c_map)
    x_space = int(len(lon) * interval / (lon[-1] - lon[0]))
    y_space = int(len(lat) * interval / (lat[-1] - lat[0]))
    plt.xticks(np.arange(0, len(lon), x_space), [
               format_longitude(lon[i]) for i in range(0, len(lon), x_space)])
    plt.yticks(np.arange(0, len(lat), y_space), [
               format_latitude(lat[i]) for i in range(0, len(lat), y_space)])
    plt.colorbar()
    plt.savefig(
        picname, bbox_inches='tight') if picname is not None else plt.show()
    plt.close()


# ** 绘制矢量场
def plot_quiver(u, v, lon, lat, picname=None, cmap='coolwarm', scale=0.25, width=0.002, x_space=5, y_space=5):
    # 创建新的网格位置变量(lat_c, lon_c)
    if len(lon.shape) == 1 and len(lat.shape) == 1:
        lon_c, lat_c = np.meshgrid(lon, lat)
    else:
        lon_c, lat_c = lon, lat

    # 设置箭头的比例、颜色、宽度等参数
    # scale = 0.25  # 箭头的大小 / 缩小程度
    # color = '#E5D1FA'
    # width = 0.002  # 箭头的宽度
    # x_space = 1
    # y_space = 1

    # 计算矢量的大小
    S = xr.DataArray(np.hypot(np.array(u), np.array(v)))

    mean_S = S.nanmean()

    # 使用 plt.quiver 函数绘制矢量图
    # 通过设置 quiver 函数的 pivot 参数来指定箭头的位置
    quiver_plot = plt.quiver(lon_c[::y_space, ::x_space],
                             lat_c[::y_space, ::x_space],
                             u[::y_space, ::x_space],
                             v[::y_space, ::x_space],
                             S[::y_space, ::x_space],  # 矢量的大小，可以不要
                             pivot='middle',
                             scale=scale,
                             #  color=color, # 矢量的颜色，单色
                             cmap=cmap,  # 矢量的颜色，多色
                             width=width)
    # plt.quiverkey(quiver_plot, X=0.90, Y=0.975, U=1, label='1 m/s', labelpos='E', fontproperties={'size': 10})
    plt.quiverkey(quiver_plot, X=0.87, Y=0.975, U=mean_S,
                  label=f'{mean_S:.2f} m/s', labelpos='E', fontproperties={'size': 10})
    plt.colorbar(quiver_plot)
    plt.xlabel('X')
    plt.ylabel('Y')

    plt.savefig(
        picname, bbox_inches='tight') if picname is not None else plt.show()
    plt.clf()
    plt.close()


def plot_contourf_cartopy(data, lon, lat, picname=None, cmap='rainbow', cn_fill_num=20, fig_size=(12, 9), title='Cartopy', land_color='green', ocean_color='lightgrey'):
    if len(lon.shape) == 2:
        lon = lon[0, :]
    if len(lat.shape) == 2:
        lat = lat[:, 0]
    
    data_max = np.nanmax(data)
    data_min = np.nanmin(data)
    levels = np.linspace(data_min, data_max, cn_fill_num)
    cbar_ticks = np.linspace(data_min, data_max, 9)

    fig = plt.figure(figsize=fig_size)
    proj = ccrs.PlateCarree()
    ax = fig.add_subplot(1, 1, 1, projection=proj)
    
    ax.set_extent([lon[0], lon[-1], lat[0], lat[-1]], crs=proj)

    ax.add_feature(cfeature.LAND, facecolor=land_color, edgecolor="k")
    ax.add_feature(cfeature.COASTLINE, linewidth=0.5)
    # ax.add_feature(cfeature.BORDERS, linewidth=0.5, linestyle=":") # 加载国界线

    ax.set_xticks(np.arange(lon[0], lon[-1] + 1e-5, 5), crs=proj)
    ax.set_yticks(np.arange(lat[0], lat[-1] + 1e-5, 5), crs=proj)
    lon_formatter = LongitudeFormatter(zero_direction_label=True)
    ax.xaxis.set_major_formatter(lon_formatter)
    lat_formatter = LatitudeFormatter()
    ax.yaxis.set_major_formatter(lat_formatter)
    # plt.title(title, fontsize=16)

    cmap = plt.get_cmap(cmap)
    X, Y = np.meshgrid(lon, lat)

    # Fill color for land and ocean
    # ax.background_patch.set_facecolor(ocean_color)
    ax.patch.set_facecolor(ocean_color)
    ax.add_feature(cfeature.LAND, facecolor=land_color)

    cticks = cbar_ticks
    norm = mpl.colors.BoundaryNorm(cticks, cmap.N)

    cnplot = ax.contourf(X, Y, data, levels=levels, cmap=cmap,
                         norm=norm, transform=proj, extend='both', alpha=1, zorder=0)
    # cllevels = np.linspace(data_min, data_max, 9)
    # clplot = ax.contour(X, Y, data, levels=levels[9::10], colors='k', linewidths=0.5, transform=proj, zorder=1, alpha=0.8, linestyle='--')
    # 添加色标，并选择位置
    divider = make_axes_locatable(ax)
    location = 3
    if location == 1:  # 左侧
        cax = divider.new_horizontal(
            size="5%", pad=1, axes_class=plt.Axes, pack_start=True)
        fig.add_axes(cax)
        cbar = plt.colorbar(
            cnplot, cax=cax, orientation='vertical', extend='both')
    elif location == 2:  # 下方
        cax = divider.new_vertical(
            size="5%", pad=0.3, axes_class=plt.Axes, pack_start=True)
        fig.add_axes(cax)
        cbar = plt.colorbar(
            cnplot, cax=cax, orientation='horizontal', extend='both')
    elif location == 3:  # 右侧
        cax = divider.new_horizontal(size="5%", pad=0.1, axes_class=plt.Axes)
        fig.add_axes(cax)
        # cbar = plt.colorbar(cnplot, cax=cax, orientation='vertical', extend='both', format='%.0f')
        cbar = fig.colorbar(mpl.cm.ScalarMappable(cmap=cmap, norm=norm),
                            cax=cax, orientation='vertical', extend='both', format='%.3f')
        cax.yaxis.set_ticks_position('right')
        cax.yaxis.set_label_position('right')
    else:  # 上方
        cax = divider.new_vertical(size="5%", pad=0.2, axes_class=plt.Axes)
        fig.add_axes(cax)
        cbar = plt.colorbar(
            cnplot, cax=cax, orientation='horizontal', extend='both')
    cbar.ax.tick_params(labelsize=10)
    cbar.ax.xaxis.set_tick_params(direction='in', width=1, length=2)
    # 添加cbar_ticks
    # cbar.set_ticks(np.arange(round(levels[0]), round(levels[-1]), round((levels[-1]-levels[0])/9)))  # 设置色标刻度
    # cbar.set_ticks(cbar_ticks)  # 设置色标刻度

    cbar.set_ticks(cticks)
    # cbar.ax.ticks = np.linspace(data_min, data_max, 8)
    # cbar.set_ticks(np.arange(round(levels[0]), round(levels[-1]), round((levels[-1]-levels[0])/9)))  # 设置色标刻度

    # 单独设置label
    cbar.set_label(title, fontsize=10,
                   weight='bold')
    # cax.set_position([0.1, 0.2, 0.02, 0.6]) # 调整色标位置
    fig.savefig(
        picname, bbox_inches='tight', dpi=600) if picname is not None else plt.show()
    plt.close()

if __name__ == '__main__':
    # ** 绘制填色图
    data = np.random.randn(100, 100)
    picname = 'test.png'
    plot_contourf(data, picname, c_map='rainbow', minmax=None,
                  labels=None, ticks_space=None, ticks=None, figsize=(12, 9))
    # ** 绘制矢量场
    u = np.random.randn(100, 100)
    v = np.random.randn(100, 100)
    lon = np.linspace(0, 360, 100)
    lat = np.linspace(-90, 90, 100)
    picname = 'test.png'
    plot_quiver(u, v, lon, lat, picname, cmap='coolwarm',
                scale=0.25, width=0.002, x_space=5, y_space=5)
    # ** 绘制经纬度填色图
    data = np.random.randn(100, 100)
    lon = np.linspace(0, 360, 100)
    lat = np.linspace(-90, 90, 100)
    picname = 'test.png'
    plot_contourf_lonlat(data, lon, lat, interval=5,
                         picname=picname, c_map='rainbow')
    # ** 制作动图
    image_list = ['test1.png', 'test2.png', 'test3.png']
    gif_name = 'test.gif'
    create_gif(image_list, gif_name, duration=0.2)



#!/usr/bin/env python
# coding=utf-8
'''
Author: Liu Kun && 16031215@qq.com
Date: 2024-09-17 17:12:47
LastEditors: Liu Kun && 16031215@qq.com
LastEditTime: 2024-10-06 19:05:38
FilePath: \\Python\\My_Funcs\\OAFuncs\\OAFuncs\\oa_data.py
Description:  
EditPlatform: vscode
ComputerInfo: XPS 15 9510
SystemInfo: Windows 11
Python Version: 3.11
'''



import multiprocessing as mp
from concurrent.futures import ThreadPoolExecutor

import numpy as np
from scipy.interpolate import griddata

__all__ = ['interp_2d', 'interp_2d_parallel']

# ** 高维插值函数，插值最后两个维度
def interp_2d(target_x, target_y, origin_x, origin_y, data, method='linear'):
    """
    高维插值函数，默认插值最后两个维度，传输数据前请确保数据的维度正确
    参数:
    target_y (array-like): 目标经度网格 1D 或 2D
    target_x (array-like): 目标纬度网格 1D 或 2D
    origin_y (array-like): 初始经度网格 1D 或 2D
    origin_x (array-like): 初始纬度网格 1D 或 2D
    data (array-like): 数据 (*, lat, lon) 2D, 3D, 4D
    method (str, optional): 插值方法，可选 'linear', 'nearest', 'cubic' 等，默认为 'linear'
    返回:
    array-like: 插值结果
    """

    # 确保目标网格和初始网格都是二维的
    if len(target_y.shape) == 1:
        target_x, target_y = np.meshgrid(target_x, target_y)
    if len(origin_y.shape) == 1:
        origin_x, origin_y = np.meshgrid(origin_x, origin_y)

    dims = data.shape
    len_dims = len(dims)
    # print(dims[-2:])
    # 根据经纬度网格判断输入数据的形状是否匹配

    if origin_x.shape != dims[-2:] or origin_y.shape != dims[-2:]:
        print(origin_x.shape, dims[-2:])
        raise ValueError(
            'Shape of data does not match shape of origin_x or origin_y.')

    # 将目标网格展平成一维数组
    target_points = np.column_stack((np.ravel(target_y), np.ravel(target_x)))

    # 将初始网格展平成一维数组
    origin_points = np.column_stack((np.ravel(origin_y), np.ravel(origin_x)))

    # 进行插值
    if len_dims == 2:
        interpolated_data = griddata(origin_points, np.ravel(
            data), target_points, method=method)
        interpolated_data = np.reshape(interpolated_data, target_y.shape)
    elif len_dims == 3:
        interpolated_data = []
        for i in range(dims[0]):
            dt = griddata(origin_points, np.ravel(
                data[i, :, :]), target_points, method=method)
            interpolated_data.append(np.reshape(dt, target_y.shape))
            print(f'Interpolating {i+1}/{dims[0]}...')
        interpolated_data = np.array(interpolated_data)
    elif len_dims == 4:
        interpolated_data = []
        for i in range(dims[0]):
            interpolated_data.append([])
            for j in range(dims[1]):
                dt = griddata(origin_points, np.ravel(
                    data[i, j, :, :]), target_points, method=method)
                interpolated_data[i].append(np.reshape(dt, target_y.shape))
                print(
                    f'\rInterpolating {i*dims[1]+j+1}/{dims[0]*dims[1]}...', end='')
        print('\n')
        interpolated_data = np.array(interpolated_data)

    return interpolated_data


# ** 高维插值函数，插值最后两个维度，使用多线程进行插值
# 在本地电脑上可以提速三倍左右，超算上暂时无法加速
def interp_2d_parallel(target_x, target_y, origin_x, origin_y, data, method='linear'):
    def interp_single2d(target_y, target_x, origin_y, origin_x, data, method='linear'):
        target_points = np.column_stack(
            (np.ravel(target_y), np.ravel(target_x)))
        origin_points = np.column_stack(
            (np.ravel(origin_y), np.ravel(origin_x)))

        dt = griddata(origin_points, np.ravel(
            data[:, :]), target_points, method=method)
        return np.reshape(dt, target_y.shape)

    def interp_single3d(i, target_y, target_x, origin_y, origin_x, data, method='linear'):
        target_points = np.column_stack(
            (np.ravel(target_y), np.ravel(target_x)))
        origin_points = np.column_stack(
            (np.ravel(origin_y), np.ravel(origin_x)))

        dt = griddata(origin_points, np.ravel(
            data[i, :, :]), target_points, method=method)
        return np.reshape(dt, target_y.shape)

    def interp_single4d(i, j, target_y, target_x, origin_y, origin_x, data, method='linear'):
        target_points = np.column_stack(
            (np.ravel(target_y), np.ravel(target_x)))
        origin_points = np.column_stack(
            (np.ravel(origin_y), np.ravel(origin_x)))

        dt = griddata(origin_points, np.ravel(
            data[i, j, :, :]), target_points, method=method)
        return np.reshape(dt, target_y.shape)

    if len(target_y.shape) == 1:
        target_x, target_y = np.meshgrid(target_x, target_y)
    if len(origin_y.shape) == 1:
        origin_x, origin_y = np.meshgrid(origin_x, origin_y)

    dims = data.shape
    len_dims = len(dims)

    if origin_x.shape != dims[-2:] or origin_y.shape != dims[-2:]:
        raise ValueError('数据形状与 origin_x 或 origin_y 的形状不匹配.')

    interpolated_data = []

    # 使用多线程进行插值
    with ThreadPoolExecutor(max_workers=mp.cpu_count()-2) as executor:
        print(f'Using {mp.cpu_count()-2} threads...')
        if len_dims == 2:
            interpolated_data = list(executor.map(interp_single2d, [target_y], [
                                     target_x], [origin_y], [origin_x], [data], [method]))
        elif len_dims == 3:
            interpolated_data = list(executor.map(interp_single3d, [i for i in range(dims[0])], [
                                     target_y]*dims[0], [target_x]*dims[0], [origin_y]*dims[0], [origin_x]*dims[0], [data]*dims[0], [method]*dims[0]))
        elif len_dims == 4:
            interpolated_data = list(executor.map(interp_single4d, [i for i in range(dims[0]) for j in range(dims[1])], [j for i in range(dims[0]) for j in range(dims[1])], [
                                     target_y]*dims[0]*dims[1], [target_x]*dims[0]*dims[1], [origin_y]*dims[0]*dims[1], [origin_x]*dims[0]*dims[1], [data]*dims[0]*dims[1], [method]*dims[0]*dims[1]))
            interpolated_data = np.array(interpolated_data).reshape(
                dims[0], dims[1], target_y.shape[0], target_x.shape[1])

    interpolated_data = np.array(interpolated_data)

    return interpolated_data


if __name__ == '__main__':
    import time

    import matplotlib.pyplot as plt

    # 测试数据
    origin_x = np.linspace(0, 10, 11)
    origin_y = np.linspace(0, 10, 11)
    target_x = np.linspace(0, 10, 101)
    target_y = np.linspace(0, 10, 101)
    data = np.random.rand(11, 11)

    # 高维插值
    origin_x = np.linspace(0, 10, 11)
    origin_y = np.linspace(0, 10, 11)
    target_x = np.linspace(0, 10, 101)
    target_y = np.linspace(0, 10, 101)
    data = np.random.rand(10, 10, 11, 11)

    start = time.time()
    interpolated_data = interp_2d(target_x, target_y, origin_x, origin_y, data)
    print(f'Interpolation time: {time.time()-start:.2f}s')

    print(interpolated_data.shape)

    # 高维插值多线程
    start = time.time()
    interpolated_data = interp_2d_parallel(
        target_x, target_y, origin_x, origin_y, data)
    print(f'Interpolation time: {time.time()-start:.2f}s')

    print(interpolated_data.shape)
    print(interpolated_data[0, 0, :, :].shape)
    plt.figure()
    plt.contourf(target_x, target_y, interpolated_data[0, 0, :, :])
    plt.colorbar()
    plt.show()



#!/usr/bin/env python
# coding=utf-8
'''
Author: Liu Kun && 16031215@qq.com
Date: 2024-09-17 16:09:20
LastEditors: Liu Kun && 16031215@qq.com
LastEditTime: 2024-10-06 19:43:24
FilePath: \\Python\\My_Funcs\\OAFuncs\\OAFuncs\\__init__.py
Description:  
EditPlatform: vscode
ComputerInfo: XPS 15 9510
SystemInfo: Windows 11
Python Version: 3.11
'''

# 会导致OAFuncs直接导入所有函数，不符合模块化设计
from .oa_cmap import *
from .oa_data import *
from .oa_draw import *
from .oa_file import *
from .oa_nc import *



#!/usr/bin/env python
# coding=utf-8
'''
Author: Liu Kun && 16031215@qq.com
Date: 2024-10-06 19:25:29
LastEditors: Liu Kun && 16031215@qq.com
LastEditTime: 2024-10-06 19:34:23
FilePath: \\Python\\My_Funcs\\OAFuncs\\OAFuncs\\oa_help.py
Description:  
EditPlatform: vscode
ComputerInfo: XPS 15 9510
SystemInfo: Windows 11
Python Version: 3.11
'''

import OAFuncs

def query(lib='OAFuncs'):
    funcs = [func for func in dir(lib) if callable(getattr(lib, func))]
    print('函数数量：')
    print(len(funcs))
    print('函数列表：')
    print(funcs)

def use(func='get_var'):
    print('模块全路径：')
    print(getattr(OAFuncs, func).__module__+'.'+func)
    print('函数提示：')
    print(getattr(OAFuncs, func).__doc__)



#!/usr/bin/env python
# coding=utf-8
'''
Author: Liu Kun && 16031215@qq.com
Date: 2024-09-17 14:58:50
LastEditors: Liu Kun && 16031215@qq.com
LastEditTime: 2024-10-06 19:11:30
FilePath: \\Python\\My_Funcs\\OAFuncs\\OAFuncs\\oa_nc.py
Description:  
EditPlatform: vscode
ComputerInfo: XPS 15 9510
SystemInfo: Windows 11
Python Version: 3.11
'''



import netCDF4 as nc
import os
import numpy as np
import xarray as xr

__all__ = ['get_var', 'extract5nc', 'write2nc']

def get_var(file, *vars):
    '''
    datas_ecm = get_var(file_ecm, 'h', 't', 'u', 'v')
    '''
    ds = xr.open_dataset(file)
    datas = []
    for var in vars:
        data = ds[var]
        datas.append(data)
    ds.close()
    return datas


def extract5nc(file, varname):
    '''
    提取nc文件中的变量，再将相应维度提取，建立字典
    返回变量及字典
    '''
    ds = xr.open_dataset(file)
    vardata = ds[varname]
    dims = vardata.dims
    dimdict = {}
    for dim in dims:
        dimdict[dim] = ds[dim].values
    ds.close()
    return np.array(vardata), dimdict


def _numpy_to_nc_type(numpy_type):
    """将NumPy数据类型映射到NetCDF数据类型"""
    numpy_to_nc = {
        'float32': 'f4',
        'float64': 'f8',
        'int8': 'i1',
        'int16': 'i2',
        'int32': 'i4',
        'int64': 'i8',
        'uint8': 'u1',
        'uint16': 'u2',
        'uint32': 'u4',
        'uint64': 'u8',
    }
    return numpy_to_nc.get(str(numpy_type), 'f4')  # 默认使用 'float32'


def write2nc(file, data, varname, coords, mode):
    '''
    file: 文件路径
    data: 数据
    varname: 变量名
    coords: 坐标，字典，键为维度名称，值为坐标数据
    mode: 写入模式，'w'为写入，'a'为追加
    '''
    # 判断mode是写入还是追加
    if mode == 'w':
        if os.path.exists(file):
            os.remove(file)
            print("Warning: File already exists. Deleting it.")
    elif mode == 'a':
        if not os.path.exists(file):
            print("Warning: File doesn't exist. Creating a new file.")
            mode = 'w'

    # 打开 NetCDF 文件
    with nc.Dataset(file, mode, format="NETCDF4") as ncfile:
        # 处理坐标
        for dim, coord_data in coords.items():
            add_coords = True
            # 判断坐标是否存在，若存在，则替换/报错
            if ncfile.dimensions:
                # 返回字典，字典、列表、元组若为空，都表示False
                if dim in ncfile.dimensions:
                    # del nc.dimensions[dim]
                    if len(coord_data) != len(ncfile.dimensions[dim]):
                        raise ValueError(
                            "Length of coordinate does not match the dimension length.")
                    else:
                        add_coords = False
                        print(
                            f"Warning: Coordinate '{dim}' already exists. Replacing it.")
                        ncfile.variables[dim][:] = np.array(coord_data)
            if add_coords:
                # 创建新坐标
                ncfile.createDimension(dim, len(coord_data))
                ncfile.createVariable(dim, _numpy_to_nc_type(
                    coord_data.dtype), (dim,))
                ncfile.variables[dim][:] = np.array(coord_data)

        # 判断变量是否存在，若存在，则删除原变量
        add_var = True
        if varname in ncfile.variables:
            print(f"Warning: Variable '{varname}' already exists.")
            if data.shape != ncfile.variables[varname].shape:
                raise ValueError(
                    "Shape of data does not match the variable shape.")
            else:
                # 写入数据
                ncfile.variables[varname][:] = data
                add_var = False
                print(
                    f"Warning: Variable '{varname}' already exists. Replacing it.")

        if add_var:
            # 创建变量及其维度
            dim_names = tuple(coords.keys())  # 使用coords传入的维度名称
            ncfile.createVariable(
                varname, _numpy_to_nc_type(data.dtype), dim_names)
            # ncfile.createVariable('data', 'f4', ('time','lev'))

            # 写入数据
            ncfile.variables[varname][:] = data

        # 判断维度是否匹配
        if len(data.shape) != len(coords):
            raise ValueError(
                "Number of dimensions does not match the data shape.")




if __name__ == '__main__':
    
    data = np.random.rand(100, 50)
    write2nc(r'test.nc', data,
         'data', {'time': np.linspace(0, 120, 100), 'lev': np.linspace(0, 120, 50)}, 'a')

